#ifndef __AVB_AECP_AEM_CONTROL_VALUE_UNITS__
#define __AVB_AECP_AEM_CONTROL_VALUE_UNITS__

#define CONTROL_LINEAR_UINT8_SIZE  	1
#define CONTROL_LINEAR_UINT8_TYPE  	(uint8_t)
#define CONTROL_LINEAR_UINT_MAX_COUNT  	(44U)

#endif //__AVB_AECP_AEM_CONTROL_VALUE_UNITS__
